# cross-validation examples
# and summary of predictive modeling methods
# for classification and regression

data(iris) # we'll use this dataset for all of the examples...

# for regression:  Petal.Length vs. others
# for classification:  Species vs. others

# step 0: cleaning (impute), feature-choosing, and clustering
# step 1: investigate reliable models with cross-validation
# lm was our linear model for regression
# lmrob for robust linear regression
# glm for logistic regression; binomial(link = "logit")
# rpart for decision trees
# randomForest for those
# nnets
# svms
# knn
# gbm for boosting trees - see its example
# lmrob, too...   mm-regression error 

set.seed(1234)  # if you want reproduceability
library(cvTools)
NUM_FOLDS <- 5
cvFolds(10, K = NUM_FOLDS, type = "consecutive")
cvFolds(10, K = NUM_FOLDS, type = "interleaved")
# in each case, $which holds the
# also ... type="random" - but more awkward than using _sample_

# iris data
data <- iris
data.train <- data  # resubstitution error
data.test <- data   # when train == test

set.seed(1234)  # if you want reproduceability

# cross-validation set-up - linear
NUM_FOLDS <- 10
errors <- vector("numeric",NUM_FOLDS)  # container for errors
cvf <- cvFolds( nrow(data), K=NUM_FOLDS, type="consecutive" )
data <- data[sample(nrow(data)),]  # randomly permutes the rows
w <- cvf$which

# run the cross-validation
for (i in 1:NUM_FOLDS)
{
  data.train <- data[w!=i,]  # training data
  data.notrain <- data[w==i,]   # non-training data
  # not really necessary, but safe...
  data.test <- data.notrain[,-3]   # removes the predicted column
  
  actual_results <- data.notrain$Petal.Length
  error <- model_and_check( data.train, data.test, actual_results )
  
  errors[i] <- error
  cat("error at fold ", i, " is ", error, "\n")
}

# plot the errors created
plot( errors, pch=19, col=4 )








# builds a model and checks the results vs. actual results...
model_and_check <- function( data.train, data.test, actual_results )
{
  # build the model
  lm_mod <- lm(Petal.Length ~ ., data=data.train )
  # get the predicted results
  predicted_results <- predict( lm_mod, newdata=data.test )
  # compute the error from predicted to actual
  error <- compute_error( predicted_results, actual_results, type="rmse" )
  # return the error
  return(error)
}



# computes the error rate for continuous values
compute_error <- function( predicted_results, actual_results, type="rmse", trim_frac=0.25 )
{
  # errors from cvTools / cost
  # mspe: mean squared prediction error
  # rmspe: root...
  # mape: mean absolute prediction error
  # tmspe: trimmed mean squared prediction error, trim=0.25 by default
  # rtmspe: root...
  spe <- (predicted_results - actual_results)^2
  ape <- abs(predicted_results - actual_results)
  mape <- mean(ape) # mean abs prediction error
  mspe <- mean(spe)

  rmspe <- sqrt(mspe) # root mean squared pred. error
  if (type=="rmse" || type=="rmspe") { return(rmspe) } # usual RMSE
  if (type=="rtmse" || type=="rtmspe") {               # trimmed error
    trim_index <- ceiling(length(spe)*(1-trim_frac))
    tspe <- sort(spe)[1:trim_index]
    tmspe <- mean(tspe)
    rtmspe <- sqrt(tmspe) # trimmed!
    return(rtmspe) 
  }
  if (type=="mape" || type=="mae") { return(mape) } # abs value
  # anything else, return rmspe
  return(rmspe)
}




# cross-validation set-up - for nnets
data <- iris
NUM_FOLDS <- 10
errors <- vector("numeric",NUM_FOLDS)  # container for errors
cvf <- cvFolds( nrow(data), K=NUM_FOLDS, type="consecutive" )
data <- data[sample(nrow(data)),]  # randomly permutes the rows
w <- cvf$which

# run the cross-validation
for (i in 1:NUM_FOLDS)
{
  data.train <- data[w!=i,-5]  # training data
  data.notrain <- data[w==i,-5]   # non-training data
  # not really necessary, but safe...
  data.test <- data.notrain[,-3]   # removes the predicted column
  
  actual_results <- data.notrain$Petal.Length
  error <- model_and_check_nn( data.train, data.test, actual_results )
  
  errors[i] <- error
  cat("error at fold ", i, " is ", error, "\n")
}

# plot the errors created
plot.new()
plot( errors, pch=19, col=2 ) # 2 for nn


# here is a model_and_check function for NNets
model_and_check_nn <- function( data.train, data.test, actual_results )
{
  #library("neuralnet")
  # build the model
  model <- neuralnet(Petal.Length ~ 
                       Sepal.Length+Sepal.Width+Petal.Width,
                     data=data.train, lifesign="full", # minimal or none
                     threshold = 0.08,
                     hidden=c(5))  # 3 neurons in the hidden layer
  plot.new()  # optional!
  plot(nn)
  # get the predicted results: here, with compute
  full_predicted_results <- compute( nn, data.test )
  predicted_results <- full_predicted_results$net.result
  # compute the error from predicted to actual
  error <- compute_error( predicted_results, actual_results, type="rmse" )
  # return the error
  return(error)
}





# k-NN methods...
# are nice, because they're simple!
library(class)

# cross-validation set-up - for knn
data <- iris
data[5] <- ifelse(data[5] == "versicolor", "versicolor", "not_vc")
names(data)[5] <- "is_versicolor"
data[5] <- as.factor(data[5])

NUM_FOLDS <- 10
errors <- vector("numeric",NUM_FOLDS)  # container for errors
cvf <- cvFolds( nrow(data), K=NUM_FOLDS, type="consecutive" )
data <- data[sample(nrow(data)),]  # randomly permutes the rows
w <- cvf$which

# run the cross-validation - species!
for (i in 1:NUM_FOLDS)
{
  data.train <- data[w!=i,]  # training data
  data.notrain <- data[w==i,]   # non-training data
  # not really necessary, but safe...
  data.test <- data.notrain[,-5]   # removes the predicted column
  
  actual_results <- data.notrain$is_versicolor
  error <- model_and_check_knn( data.train, data.test, actual_results )
  
  errors[i] <- error
  cat("error at fold ", i, " is ", error, "\n")
}

# plot the errors created
plot.new()
plot( errors, pch=19, col=3 ) # 3 for knn


# here is a model_and_check function for k-nn
model_and_check_knn <- function( data.train, data.test, actual_results )
{
  #library("neuralnet")
  # build the model
  predicted_results <- knn(data.train[,-5], data.test, data.train[,5], k=5)
  # compute the error from predicted to actual
  table(predicted_results, actual_results)
  error <- sum(predicted_results != actual_results)/length(predicted_results)
  # return the error
  return(error)
}



# SVM
library(kernlab)

# cross-validation set-up - for SVM is the same as knn...
data <- iris
data[5] <- ifelse(data[5] == "versicolor", "versicolor", "not_vc")
names(data)[5] <- "is_versicolor"
data[5] <- as.factor(data[5])

NUM_FOLDS <- 10
errors <- vector("numeric",NUM_FOLDS)  # container for errors
cvf <- cvFolds( nrow(data), K=NUM_FOLDS, type="consecutive" )
data <- data[sample(nrow(data)),]  # randomly permutes the rows
w <- cvf$which

# run the cross-validation - species!
for (i in 1:NUM_FOLDS)
{
  data.train <- data[w!=i,]  # training data
  data.notrain <- data[w==i,]   # non-training data
  # not really necessary, but safe...
  data.test <- data.notrain[,-5]   # removes the predicted column
  
  actual_results <- data.notrain$is_versicolor
  error <- model_and_check_svm( data.train, data.test, actual_results )
  
  errors[i] <- error
  cat("error at fold ", i, " is ", error, "\n")
}

# plot the errors created
plot.new()
plot( errors, pch=19, col=1 ) # 1 for svm


# here is a model_and_check function for k-nn
model_and_check_svm <- function( data.train, data.test, actual_results )
{
  #library("neuralnet")
  # build the model
  model <- ksvm(data.train[,-5],data.train[,5],type="C-svc",
                       kernel='rbf',kpar=list(sigma=1),C=1)
  # predict
  predicted_results <- predict(model,newdata=data.test)
  # compute the error from predicted to actual
  #table(predicted_results, actual_results)
  error <- sum(predicted_results != actual_results)/length(predicted_results)
  # return the error
  return(error)
}












