#install.packages("gbm")
library(gbm)

# yeah!
data(iris)

# versicolor will be in red...
plot( iris$Petal.Length, iris$Petal.Width, col=as.factor(iris$Species) )


# use gbm (generalized boosting models)
model <- gbm(Species ~ Sepal.Length + Sepal.Width +
             + Petal.Length + Petal.Width, data=iris, 
             n.trees=1000, 
             interaction.depth=2,
             distribution="multinomial")  
# "bernoulli for 2 options   # "gaussian" for continuous

# summary - provides the importance of each factor...
summary(model)
# could re-run the model...


# predict!
prediction <- predict.gbm(model, newdata=iris2[45:55,], 
                          type="response", 
                          n.trees=1000)


# resub.
all_pred <- predict.gbm(model, iris2, 
                          type="response", 
                          n.trees=1000)


# take a look ...



# a copy...
iris2 <- iris

# is it versicolor - the trickiest one
newcol = data.frame(isVersicolor=(iris2$Species=='versicolor'))

# bind it to iris2
iris2 <- cbind(iris2, newcol)  # two-outcome column

attach(iris2)

# use gbm (generalized boosting models)
model <- gbm(isVersicolor ~ Sepal.Length + Sepal.Width +
               + Petal.Length + Petal.Width, data=iris, 
             n.trees=1000, 
             interaction.depth=2,
             distribution="bernoulli")  
# "bernoulli for 2 options   # "gaussian" for continuous
best.iter <- gbm.perf(model,method="OOB")
print(best.iter)


# summary - provides the importance of each factor...
summary(model)
# could re-run the model...


# predict!
prediction <- predict.gbm(model, iris2[45:55,], 
                          type="response", 
                          n.trees=1000)


# resubstitution...
prediction <- predict.gbm(model, iris2, 
                          type="response", 
                          n.trees=1000)
pred <- prediction > 0.5
table(iris2$isVersicolor,pred)





# another, larger  set of examples...
# this shows off gbm's ability to "tease out" important factors...

set.seed(1234)
N <- 1000
X1 <- runif(N)
X2 <- 2*runif(N)
X3 <- ordered(sample(letters[1:4],N,replace=TRUE),levels=letters[4:1])
X4 <- factor(sample(letters[1:6],N,replace=TRUE))
X5 <- factor(sample(letters[1:3],N,replace=TRUE))
X6 <- 3*runif(N) 
mu <- c(-1,0,1,2)[as.numeric(X3)]

SNR <- 10 # signal-to-noise ratio
Y <- .5*(X1**2) + 2 * (X2**.5) + mu

# NOTICE: it depends on X1, X2, and X3

sigma <- sqrt(var(Y)/SNR)
Y <- Y + rnorm(N,0,sigma)

# introduce some missing values
X1[sample(1:N,size=500)] <- NA
X4[sample(1:N,size=300)] <- NA

data <- data.frame(Y=Y,X1=X1,X2=X2,X3=X3,X4=X4,X5=X5,X6=X6)

# fit initial model
gbm1 <-
  gbm(Y~X1+X2+X3+X4+X5+X6,         # formula
      data=data,                   # dataset
      var.monotone=c(0,0,0,0,0,0), # -1: monotone decrease,
      distribution="gaussian",     # see the help for other choices
      n.trees=1000,                # number of trees
      shrinkage=0.05,              # shrinkage or learning rate,
      interaction.depth=3,         # 1: additive model, 2: two-way interactions, etc.
      bag.fraction = 0.5,          # subsampling fraction, 0.5 is probably best
      train.fraction = 0.5,        # fraction of data for training,
      n.minobsinnode = 10,         # minimum total weight needed in each node
      cv.folds = 3,                # do 3-fold cross-validation
      keep.data=TRUE,              # keep a copy of the dataset with the object
      verbose=T)                   # print out progress


# look at what's worthwhile...
summary(gbm1)


# check performance using an out-of-bag estimator
# OOB underestimates the optimal number of iterations
best.iter <- gbm.perf(gbm1,method="OOB")
print(best.iter)

# check performance using a 50% heldout test set
best.iter <- gbm.perf(gbm1,method="test")
print(best.iter)

# check performance using 5-fold cross-validation
best.iter <- gbm.perf(gbm1,method="cv")
print(best.iter)

# plot the performance # plot variable influence
summary(gbm1,n.trees=1)         # based on the first tree
summary(gbm1,n.trees=best.iter) # based on the estimated best number of trees

# compactly print the first and last trees for curiosity
print(pretty.gbm.tree(gbm1,1))
print(pretty.gbm.tree(gbm1,gbm1$n.trees))

# make some new data
N <- 1000
X1 <- runif(N)
X2 <- 2*runif(N)
X3 <- ordered(sample(letters[1:4],N,replace=TRUE))
X4 <- factor(sample(letters[1:6],N,replace=TRUE))
X5 <- factor(sample(letters[1:3],N,replace=TRUE))
X6 <- 3*runif(N) 
mu <- c(-1,0,1,2)[as.numeric(X3)]

Y <- .5*(X1**2) + 2 * (X2**.5) + mu + rnorm(N,0,sigma)

data2 <- data.frame(Y=Y,X1=X1,X2=X2,X3=X3,X4=X4,X5=X5,X6=X6)

# predict on the new data using "best" number of trees
# f.predict generally will be on the canonical scale (logit,log,etc.)
f.predict <- predict(gbm1,data2,best.iter)

# least squares error
print(sum((data2$Y-f.predict)^2))

# create marginal plots
# plot variable X1,X2,X3 after "best" iterations
par(mfrow=c(1,3))
plot(gbm1,1,best.iter)
plot(gbm1,2,best.iter)
plot(gbm1,3,best.iter)
par(mfrow=c(1,1))
# contour plot of variables 1 and 2 after "best" iterations
plot(gbm1,1:2,best.iter)
# lattice plot of variables 2 and 3
plot(gbm1,2:3,best.iter)
# lattice plot of variables 3 and 4
plot(gbm1,3:4,best.iter)

# 3-way plots
plot(gbm1,c(1,2,6),best.iter,cont=20)
plot(gbm1,1:3,best.iter)
plot(gbm1,2:4,best.iter)
plot(gbm1,3:5,best.iter)


